module get_1_outerm_m

  implicit none

contains

  type(ssh_contour) function get_1_outerm(cyclone, coord_extr, &
       innermost_level, outside_points, ssh, corner, step)

    ! Gets one outermost contour, if it can find one. Method:
    ! dichotomy on level of ssh. If the procedure cannot find an
    ! outermost contour, it returns a null ssh_contour. An outermost
    ! contour is not found if, and only if, there is no good contour
    ! at innermost level. If a contour is found then its level is
    ! farther from the extremum than innermost level.

    ! The length of the interval of longitudes of the domain, step(1)
    ! * (size(ssh, 1) - 1), should be < 180 degrees, so that the
    ! geometrical processing done here is non-ambiguous. The
    ! longitudes of outside points and the target extremum should be
    ! in the interval [corner(1), corner(1) + step(1) * (size(ssh, 1)
    ! - 1)] as there will be no automatic shifting by a mulitple of
    ! 360 degrees.

    ! Libraries:
    use contour_531, only: polyline
    use nr_util, only: assert

    use derived_types, only: ssh_contour, missing_ssh
    use good_contour_m, only: good_contour
    use spher_polyline_area_m, only: spher_polyline_area

    logical, intent(in):: cyclone
    real, intent(in):: coord_extr(:) ! (2)

    real, intent(in):: innermost_level
    ! ssh level of the innermost contour to consider around the target
    ! extremum, in m. Assume: innermost_level < extremum for a
    ! maximum, > extremum for a minimum.

    real, intent(in):: outside_points(:, :) ! (2, :) longitude and
    ! latitude, in rad, of all the significant extrema, except the
    ! target extremum

    real, intent(in):: ssh(:, :) ! in m

    real, intent(in):: corner(:) ! (2)
    ! longitude and latitude corresponding to ssh(1,1), in rad

    real, intent(in):: step(:) ! (2)
    ! longitude and latitude steps, in rad

    ! Local:

    real level_try, level_good, level_bad ! in m
    type(polyline) tentative_contour
    real, parameter:: precision = 5e-4 ! in m
    logical mask(size(ssh, 1), size(ssh, 2))

    !-----------------------------------------------------------------

    get_1_outerm%polyline = good_contour(corner, step, ssh, innermost_level, &
         coord_extr, outside_points)

    if (get_1_outerm%n_points == 0) then
       ! null ssh contour
       get_1_outerm%ssh = missing_ssh
       get_1_outerm%area = - 1e6
    else
       level_good = innermost_level
       mask = ssh /= huge(0.)
       level_try = merge(maxval(ssh, mask), minval(ssh, mask), cyclone)
       call assert(merge(level_try > level_good, level_try < level_good, &
            cyclone), "get_1_outerm level_try")
       tentative_contour = good_contour(corner, step, ssh, level_try, &
            coord_extr, outside_points)

       if (tentative_contour%n_points /= 0) then
          ! This can only happen if there are degenerate extrema
          ! everywhere around the path of contour. So it should
          ! probably never happen.
          get_1_outerm%polyline = tentative_contour
          get_1_outerm%ssh = level_try
       else
          ! Assertion: {no good contour at maxval or minval}
          level_bad = level_try

          do while (abs(level_bad - level_good) > precision)
             level_try = (level_good + level_bad) / 2.
             tentative_contour = good_contour(corner, step, ssh, level_try, &
                  coord_extr, outside_points)
             if (tentative_contour%n_points /= 0) then
                level_good = level_try
                get_1_outerm%polyline = tentative_contour
             else
                level_bad = level_try
             end if

             ! Assertion: get_1_outerm%polyline == good_contour(. . .,
             ! level_good) and get_1_outerm%%n_points /= 0 and
             ! no good contour at level_bad
          end do
          
          get_1_outerm%ssh = level_good
       end if

       get_1_outerm%area = spher_polyline_area(get_1_outerm%polyline)
    end if

  end function get_1_outerm

end module get_1_outerm_m
